#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
import logging

from pylab import *
from ReadRawFile import *
from datetime import date
import netCDF4
from scipy import spatial
from mpl_toolkits.basemap import pyproj
import itertools as it
import time

##-------------------------------##
#     Class for satellite name    #
##-------------------------------##
class SatName: 
    """ ### Defined a number for each satellite """
    def __init__(self):
        self.satname = array(['AMSUBN15','AMSUBN16','MHSN18','MHSN19','MHSM01','MHSM02'])
        self.satnum  = arange(size(self.satname))+1

    def get_satnum(self,name):
        return concatenate([self.satnum[nn == self.satname] for nn in name])
    
    def get_satname(self,num):
        return concatenate([self.satname[nn == self.satnum] for nn in num])

##-------------------------------##
#     Correction for one orbit    #
##-------------------------------##
def applycorr2orb(fname,lonvec,latvec,ch=['All'],area=[-35.,35.,-180.,180.]):
    """ ### Apply correction to AMSU or MHS orbit
    ## Inputs
    # fname  = file name
    # tstep  = time resolution for new grid in minutes
    # lonvec = longitude vecteur of new grid
    # latvec = latitude vecteur of new grid
    # ch     = list of extrated channel ('B1','B2','B3','B4','B5' or 'All'), default 'All'
    # area   = limits [LS,LN,LW,LE] for extracted area, default is world (ie [-35.,35.,-180.,180.])
    #
    # Outputs
    # tbcorrm = brightness temperature of orbit in new grid
    # trefm   = ordinal time 
    #
    """

    t=time.time()

    logger = logging.getLogger()

    # ## Read orbit file
    # print 'Read ', fname
    lon, lat, dt, tb, fov = load_amsub_mhs(fname,ch,area)

    try:
        # ##  Class of each orbit points
        # print 'Classification'
        clz = find_class(lat,lon)
        nozero = (clz != 0)

    except ValueError:
        logger.error('Orbit is out of the area: {0}'.format(fname))
        return [], [], []

    else:
        # ## Coef of correction as a function of classification, fov and day of year
        # print 'Correction'
        coefz = find_corrcoef(dt[nozero],fov[nozero],clz[nozero],ch=ch)
        tbcorr = array(tb)[:,nozero] - coefz
        # ## Average on new grid
        tbcorrm, tobsm = newgrid(lonvec,latvec,lon[nozero],lat[nozero],dt[nozero],tbcorr)
    
    logger.info('Deal {0} in: {1}'.format(fname, time.time()-t))

    return tbcorrm, tobsm, fname.split('/')[3]

##----------------------------------##
#      Read AMSU-B and MHS files     #
##----------------------------------##
def load_amsub_mhs(fname,ch,area):
    """ ### Read AMSU-B and MHS files
    ## Inputs
    # fname = file name
    # ch    = list of extrated channel ('B1','B2','B3','B4','B5' or 'All'), default 'All'
    # area  = limits [LS,LN,LW,LE] for extracted area, default is world (ie [-35,35,-180.,180])
    #
    ## Outputs
    # lon  = Longitude
    # lat  = Latitude
    # yday = Day of year
    # tms  = UTC time in milliseconds
    # tb   = Brightness temperature for each channel
    # err  = 1 if reading error, 0 else
    #
    """

    # Index of channel
    ch=[cch.lower() for cch in ch]
    if (ch[0] == 'all'):
        ich = arange(5)
    else:
        ich = array([int(cch[1])-1  for cch in ch])

    # Variables
    dset_names = ['amb1c_latlon','amb1c_btemps','amb1c_scnlindy','amb1c_scnlintime','amb1c_scnlinyr']
    try:
        headers, datasets = ReadRawFile(fname,'/home/mleduc/python/ReadRawFile/data/amsub-mhs_1c_header.txt','/home/mleduc/python/ReadRawFile/data/amsub-mhs_1c_record.txt', False,dset_names)   
    except:
        logger.error('Error with {0}'.format(fname))
        # continue
        return 

    # Latitude
    lat = float32(datasets['amb1c_latlon'][:,:,0]/1.E4)
    # Longitude
    lon = float32(datasets['amb1c_latlon'][:,:,1]/1.E4)
    # Year
    yyyy= datasets['amb1c_scnlinyr']
    # Day of Year
    yday= datasets['amb1c_scnlindy']
    # Time in milliseconds
    tms = datasets['amb1c_scnlintime']
    # Tb
    tb  = float32(datasets['amb1c_btemps'][:,:,ich]/1.E2)
    # Remplace fill values -9999.99 by NaN
    tb[(tb == -9999.99)] = NaN

    ### Extract Area
    # Index in the defined area
    LS, LN, LW, LE = area
    zone = (lon>=LW) & (lon<=LE) & (lat>=LS) & (lat<=LN)
    nscn, nfov, nch = tb.shape
    # Extraction of Tb for each channel
    tb = [(tb[:,:,ii])[zone] for ii in xrange(nch)]
    # day of year and time
    yyyy = transpose([yyyy]*nfov)[zone]
    yday = transpose([yday]*nfov)[zone]
    tms  = transpose([tms]*nfov)[zone]
    # Find location in fov
    fov = array(range(nfov)*nscn).reshape(nscn,nfov)[zone]+1

    ### Convert in python date/time
    dd = num2date(yday)
    tt = [ms2time(ms) for ms in tms]
    dt = array([d.replace(year=y,hour=h,minute=m,second=s) for d,y,(h,m,s) in zip(dd,yyyy,tt)])

    return lon[zone], lat[zone], dt, tb, fov

##-----------------------------------------------------##
#      Convert milliseconds in hour, minute, second     #
##-----------------------------------------------------##
def ms2time(ms):
    """ ### Convert milliseconds in hour, minute, second
    # Warning 1: lost millisecond because second is integer 2) no day
    # Warning 2: no day
    ## Inputs
    # ms  = millisecond
    # 
    ## Output
    # h = hour
    # m = minute
    # s = second
    #
    """
    s    = ms/1000
    m, s = divmod(s,60)
    h, m = divmod(m,60)
    #d, h = divmod(h,24)
    return h, m, s

##---------------------##
#      Find classes     #
##---------------------##
def find_class(lat,lon):
    """ ### Find class of each given (lat,lon) 
    ## Inputs
    # lat  = Latitude (degrees)
    # lon  = Longitude (degrees)
    # 
    ## Output
    # clpt = class of each point (same shape as lat and lon)
    #
    """

    # Read the classification file
    fic_class = '/homedata/mleduc/CorrTb/classifERAI_Trop.nc'
    fileID = netCDF4.Dataset(fic_class)
    lat_class = fileID.variables['lat'][:]
    lon_class = fileID.variables['lon'][:]
    classif   = int16(fileID.variables['classif_tot'][:])
    fileID.close()

    # Latitude limits of classification (ie tropical band)
    latstep = abs(diff(lat_class))[0]
    latmin = lat_class.min()-latstep/2.
    latmax = lat_class.max()+latstep/2.
    # Only points in the area of classification
    inclass = logical_and(lat>=latmin,lat<=latmax)

    # Geographic projection (lon/lat in degrees to x/y in m)
    projG = pyproj.Proj("+proj=merc +ellps=WGS84 +datum=WGS84")
    longd_class,latgd_class = meshgrid(lon_class,lat_class)
    xclass,yclass = projG(longd_class,latgd_class)
    xx,yy = projG(lon[inclass],lat[inclass]) 

    # Find the nearest neighbour (euclidian distance) with a kd-tree 
    tree = spatial.cKDTree(zip(ravel(xclass),ravel(yclass)))
    _, ind = tree.query(zip(xx,yy))
    clpt = zeros(shape(lat),dtype=int16)
    clpt[inclass] = classif.flatten()[ind]

    return clpt

##-------------------------------------------##
#      Find the coefficient of correction     #
##-------------------------------------------##
def find_corrcoef(dt,fov,cl,ch=['All']):
    """ ### Find the coefficient of correction
    ## Inputs
    # dt  = Python date
    # fov = Location in the swath
    # cl  = Class
    # ch  = list of extrated channel ('B1','B2','B3','B4','B5' or 'All'), default 'All'
    #
    ## Outputs
    # coef = Coefficient of correction
    #
    """
    # Index of channel
    ch=[cch.lower() for cch in ch]
    if (ch[0] == 'all'):
        ich = arange(5)
    else:
        ich = array([int(cch[1])-1  for cch in ch])
    nch = ich.size
    
    # Read the classification file
    fic_corr = '/homedata/mleduc/CorrTb/corrTb_AMSUB_MHS_L1C.nc'
    fileID = netCDF4.Dataset(fic_corr)
    day_corr  = fileID.variables['day'][:]
    coef_corr = fileID.variables['coefcorr'][:,:,:,ich]#[:]
    fileID.close()

    # Find the day of year
    yday = array([d.timetuple().tm_yday for d in dt])
    # Find the day of correction
    nday = day_corr.size
    daystep = diff(day_corr)[0]
    icorr = list(flatten([[ii]*daystep for ii in xrange(nday)])) + [0]*6
    iday = array(icorr)[yday-1]

    # Find correction
    noclass = (cl == 0)
    icl  = array(cl-1,dtype=int)
    ifov = array(fov-1,dtype=int)
    coef = []
    for ii in ich:
        dd = zeros(cl.shape)+NaN
        dd[~noclass] = coef_corr[iday[~noclass],icl[~noclass],ifov[~noclass],ii]
        coef.append(dd)

    return float32(coef)

##---------------------------##
#      New grid and time      #
##---------------------------##
def newgrid(lonvec,latvec,lon,lat,dt,tb):
    """ ### Find the coefficient of correction
    ## Inputs
    #
    #
    ## Outputs
    #
    #
    """

    # from date to ordinal
    dtord = date2num(dt)

    ### Find index of each pixel in new grid  
    # Geographic projection (lon/lat in degrees to x/y in m)
    projG = pyproj.Proj("+proj=merc +ellps=WGS84 +datum=WGS84")
    longd, latgd = meshgrid(lonvec,latvec)
    xgd, ygd = projG(longd,latgd)
    xx, yy = projG(lon,lat) 
    # Find the nearest neighbour (euclidian distance) with a kd-tree 
    tree = spatial.cKDTree(zip(ravel(xgd),ravel(ygd)))
    _, indll = tree.query(zip(xx,yy))

    # Mean
    nch = tb.shape[0]
    tbm = zeros((nch,lonvec.size*latvec.size),dtype=float32)+NaN
    dtm = zeros((lonvec.size*latvec.size))+NaN
    for ii in unique(indll):
        tbm[:,ii] = nanmean(tb[:,(indll==ii)],axis=1)
        dtm[ii]  = nanmean(dtord[(indll==ii)])

    return tbm, dtm
